module TalltyImportExport
  class Export
    attr_reader :klass

    def initialize klass
      @klass = klass
    end

    # [
    #   { key: 'entry_user_name', name: '被考核人', chain: [:entry, :user_name] },
    #   { key: 'entry_user_code', name: '被考核人工号' },
    #   { key: 'entry_user_department_name', name: '被考核人部门' },
    #   { key: 'dimension_name', name: '考核维度' },
    #   { key: 'user_name', name: '考核人' },
    #   { key: 'user_code', name: '考核人工号' },
    #   { key: 'user_department_name', name: '考核人部门' },
    #   { key: 'state', name: '考核状态', method: :state_zh },
    #   { key: 'score', name: '考核分' },
    # ]
    # export_headers_result / headers
    # key: 属性的英文名，可以支持user.name这样的方式
    # name: 属性的中文名
    # attr_type: 属性的类型
    # format: excel是否需要特定的格式，目前主要是类似于身份证号，可以用string
    # method: 导出时本地调用的方法
    # chain: 导出时对象属性通过链式调用
    # index: 数组方式，需要嵌套拿到里面的
    # merge: true/false，默认false，某一列，如果上下行的内容相同，则直接合并单元格
    # json: model_payload，代表存储在某一个列中

    def export_xlsx records, **options
      records = with_scope records
      process_options(options)

      options = options.with_indifferent_access

      Axlsx::Package.new do |pack|
        workbook = pack.workbook

        if @group_by.present?
          if records.is_a?(Array)
            records.group_by { |record| record.send(@group_by)}.each do |key, group_records|
              @group_key = key
              export_workbook workbook, group_records
            end
          else
            records.group(@group_by).count.keys.each do |key|
              @group_key = key
              export_workbook workbook, records.ransack("#{@group_where}" => key).result
            end
          end
        else
          export_workbook workbook, records
        end

        file_path = File.join(Rails.root, 'public', 'export')
        FileUtils.mkdir_p(file_path) unless Dir.exist?(file_path)
        file_name = "#{Time.now.strftime('%Y%m%d%H%M%S')}#{@filename}.xlsx"
        file_path_with_name = File.join(file_path, file_name)
        pack.serialize(file_path_with_name)
        return file_path_with_name
      end
    end

    def export_workbook workbook, records
      # excel导出样式
      alignment = { vertical: :center, horizontal: :center }
      border    = { color: '969696', style: :thin }
      title1 = workbook.styles.add_style(alignment: alignment, border: border, sz: 12, b: true)
      title2 = workbook.styles.add_style(alignment: alignment, border: border, bg_color: "2a5caa",  sz: 12, fg_color: "fffffb")
      title3 = workbook.styles.add_style(alignment: alignment.merge(wrap_text: true), border: border, sz: 10)
      headers = export_headers_result

      _sheet_name = respond_to?(:sheet_name) ? self.sheet_name : nil

      workbook.add_worksheet(name: _sheet_name) do |sheet|
        if respond_to?(:first_header)
          row_index = Axlsx.col_ref(headers.size - 1)
          sheet.merge_cells("A1:#{row_index}1")
          sheet.add_row [first_header], style: title1, height: 40
        end

        sheet.add_row headers.map{|header| header[:name]}, style: title2, height: 39

        last_row = nil
        merge_column_hash = {}
        first_content_row_index = respond_to?(:first_header) ? 2 : 1

        each_method = records.is_a?(Array) ? :each : :each
        records.send(each_method).with_index do |record, index|
          row = []
          headers.each_with_index do |header, col_index|
            _data = handle_data(record, header, index)
            if header[:merge].present? && last_row.present? && _data == last_row[col_index]
              # 这里使用二维数组，每个数组里都是列内容相同的各行
              merge_column_hash[col_index] ||= []
              if merge_column_hash[col_index].last&.last == index + first_content_row_index - 1
                # 说明内容和上面的是延续的，继续加入之前的数组
                merge_column_hash[col_index].last << index + first_content_row_index
              else
                merge_column_hash[col_index] << [index + first_content_row_index - 1, index + first_content_row_index]
              end
            end
            row.push(_data)
          end
          sheet.add_row row, style: title3, height: @row_height, types: headers.map{|header| header[:format]&.to_sym}
          last_row = row
        end
        # 需要根据column进行多行的内容合并
        if merge_column_hash.present?
          merge_column_hash.each do |col_index, row_arr|
            row_arr.each do |arr|
              sheet.merge_cells(
                Axlsx::cell_r(col_index, arr.first) + ':' + Axlsx::cell_r(col_index, arr.last)
              )
            end
          end
        end
        sheet.column_widths(*headers.map{|header| (header[:width] || @width).to_f})
      end
    end

    def process_options options = {}
      options = options.with_indifferent_access

      @row_height ||= options.delete(:row_height) || 35
      @width ||= (options.delete(:width) || 20).to_f
      @filename ||= options.delete(:filename)
      @group_by ||= options.delete(:group_by)
      @group_where = "#{@group_by}_eq" if @group_by.present?
      @headers ||= options.delete(:headers)
    end

    def with_scope records
      records
    end

    def export_headers_result
      @headers || export_headers
    end

    def export_headers **args
      @headers || klass.try(:headers) || klass.try(:model_headers)
    end

    # 处理一个记录的数据
    def handle_data record, header, index=0
      data =
        if header[:key] == '_index'
          index + 1
        elsif header[:method].present?
          send(header[:method], record, header)
        elsif header[:chain].present?
          try_chain(record, header[:chain])
        elsif header[:json]
          record.send(header[:json])[header[:key]]
        else
          try_method(record, header[:key])
        end
        handle_format(data, header)
    rescue
      ''
    end

    def try_chain record, arr
      arr.inject(record, :try)
    end

    def try_method record, method
      arr = method.to_s.split(/\./)
      try_chain record, arr
    end

    # 根据数据类型 attr_type 进行数据的格式化
    def handle_format data, header
      case header[:attr_type].to_s
      when 'string'
        data.to_s
      when 'datetime'
        data ? data.strftime('%F %H:%M') : nil
      when 'date'
        data ? data.strftime('%F') : nil
      else
        data
      end
    end
  end
end
